/*
	LeadFilter.cpp - GPS lag remover for Arduino
	Original Code by Jason Short. DIYDrones.com, changes by Andras Schaffer

	This library is free software; you can redistribute it and / or
		modify it under the terms of the GNU Lesser General Public
		License as published by the Free Software Foundation; either
		version 2.1 of the License, or (at your option) any later version.

*/

#include "Arduino.h"
#include <math.h>
#include "filter.h"

int32_t LeadFilter::get_position(int32_t pos, int16_t vel, float lag_in_seconds)
{
    int16_t accel_contribution = (vel - _last_velocity) * lag_in_seconds * lag_in_seconds;
    int16_t vel_contribution = vel * lag_in_seconds;

    // store velocity for next iteration
    _last_velocity = vel;

    return pos + vel_contribution + accel_contribution;
}

int32_t PICtrl::get_p(int32_t error)
{
	return (float)error * _kp;
}

int32_t PICtrl::get_i(int32_t error, float dt)
{
	if(dt != 0){
		_integrator += ((float)error * _ki) * dt;

		if (_integrator < -_imax) {
			_integrator = -_imax;
		} else if (_integrator > _imax) {
			_integrator = _imax;
		}
	}
	return _integrator;
}

int32_t PICtrl::get_pi(int32_t error, float dt)
{
	return get_p(error) + get_i(error, dt);
}

void
PICtrl::reset_I()
{
	_integrator = 0;
}

int32_t PIDCtrl::get_p(int32_t error)
{
	return (float)error * _kp;
}

int32_t PIDCtrl::get_i(int32_t error, float dt)
{
	if((_ki != 0) && (dt != 0)){
		_integrator += ((float)error * _ki) * dt;
		if (_integrator < -_imax) {
			_integrator = -_imax;
		} else if (_integrator > _imax) {
			_integrator = _imax;
		}
		return _integrator;
	}
	return 0;
}

int32_t PIDCtrl::get_d(int32_t input, float dt)
{
	if ((_kd != 0) && (dt != 0)) {
		_derivative = (input - _last_input) / dt;

		// discrete low pass filter, cuts out the
		// high frequency noise that can drive the controller crazy
		_derivative = _last_derivative +
		        (dt / ( _filter + dt)) * (_derivative - _last_derivative);

		// update state
		_last_input 		= input;
		_last_derivative    = _derivative;

		// add in derivative component
		return _kd * _derivative;
	}
	return 0;
}

int32_t PIDCtrl::get_pi(int32_t error, float dt)
{
	return get_p(error) + get_i(error, dt);
}


int32_t PIDCtrl::get_pid(int32_t error, float dt)
{
	return get_p(error) + get_i(error, dt) + get_d(error, dt);
}

void
PIDCtrl::reset_I()
{
	_integrator = 0;
	_last_input = 0;
	_last_derivative = 0;
}
