#ifndef Filter_h
#define Filter_h

#include "Arduino.h"
#undef abs
#undef round
#include <stdint.h>
#include <math.h>		// for fabs()

class LeadFilter {
public:
    LeadFilter() :
        _last_velocity(0) {
    }

    // setup min and max radio values in CLI
    int32_t         get_position(int32_t pos, int16_t vel, float lag_in_seconds = 1.0);
    void            clear() { _last_velocity = 0; }

private:
    int16_t         _last_velocity;

};

class PIDCtrl {
public:

	PIDCtrl(
	    const float &initial_p = 0.0,
	    const float &initial_i = 0.0,
	    const float &initial_d = 0.0,
	    const int16_t &initial_imax = 0.0) :

		_kp  (initial_p),
		_ki  (initial_i),
		_kd  (initial_d),
		_imax(initial_imax)
	{
		// no need for explicit load
	}

	/// Iterate the PID, return the new control value
	///
	/// Positive error produces positive output.
	///
	/// @param error	The measured error value
	/// @param dt		The time delta in milliseconds (note
	///					that update interval cannot be more
	///					than 65.535 seconds due to limited range
	///					of the data type).
	/// @param scaler	An arbitrary scale factor
	///
	/// @returns		The updated control output.
	///
	int32_t get_pid(int32_t error, float dt);
	int32_t get_pi(int32_t error, float dt);
	int32_t get_p(int32_t error);
	int32_t get_i(int32_t error, float dt);
	int32_t get_d(int32_t error, float dt);


	/// Reset the PID integrator
	///
	void	reset_I();

	/// Overload the function call operator to permit relatively easy initialisation
	void operator() (const float p,
	                 const float i,
	                 const float d,
	                 const int16_t imaxval) {
		_kp = p; _ki = i; _kd = d; _imax = abs(imaxval);
	}

	float	kP() const				{ return _kp; }
	float	kI() const 				{ return _ki; }
	float	kD() const 				{ return _kd; }
	int16_t	imax() const			{ return _imax; }

	void	kP(const float v)		{ _kp = v; }
	void	kI(const float v)		{ _ki = v; }
	void	kD(const float v)		{ _kd = v; }
	void	imax(const int16_t v)	{ _imax = abs(v); }

	float	get_integrator() const	{ return _integrator; }
	void	set_integrator(float i)	{ _integrator = i; }


private:
	float			_kp;
	float			_ki;
	float			_kd;
	int16_t  		_imax;

	float				_integrator;		///< integrator value
	int32_t				_last_input;		///< last input for derivative
	float				_last_derivative; 	///< last derivative for low-pass filter
	float				_output;
	float				_derivative;

	/// Low pass filter cut frequency for derivative calculation.
	///
	static const float _filter = 7.9577e-3; // Set to  "1 / ( 2 * PI * f_cut )";
	// Examples for _filter:
	// f_cut = 10 Hz -> _filter = 15.9155e-3
	// f_cut = 15 Hz -> _filter = 10.6103e-3
	// f_cut = 20 Hz -> _filter =  7.9577e-3
	// f_cut = 25 Hz -> _filter =  6.3662e-3
	// f_cut = 30 Hz -> _filter =  5.3052e-3
};

/// @class	PICtrl
/// @brief	Object managing one PI control
class PICtrl {
public:

	/// Constructor for PI that saves its settings to EEPROM
	///
	/// @note	PI must be named to avoid either multiple parameters with the
	///			same name, or an overly complex constructor.
	///
	/// @param  initial_p       Initial value for the P term.
    /// @param  initial_i       Initial value for the I term.
    /// @param  initial_imax    Initial value for the imax term.4
	///
	PICtrl(const float &initial_p = 0.0,
		   const float &initial_i = 0.0,
		   const int16_t &initial_imax = 0.0) :
		_kp  (initial_p),
		_ki  (initial_i),
		_imax(initial_imax)
	{
		// no need for explicit load, assuming that the main code uses AP_Var::load_all.
	}

	/// Iterate the PI, return the new control value
	///
	/// Positive error produces positive output.
	///
	/// @param error	The measured error value
	/// @param dt		The time delta in milliseconds (note
	///					that update interval cannot be more
	///					than 65.535 seconds due to limited range
	///					of the data type).
	/// @param scaler	An arbitrary scale factor
	///
	/// @returns		The updated control output.
	///
	//long 	get_pi(int32_t error, float	 dt);
	int32_t get_pi(int32_t error, float dt);
	int32_t get_p(int32_t error);
	int32_t get_i(int32_t error, float dt);


	/// Reset the PI integrator
	///
	void	reset_I();


	/// @name	parameter accessors
	//@{

	/// Overload the function call operator to permit relatively easy initialisation
	void operator() (const float p,
	                 const float i,
	                 const int16_t imaxval) {
		_kp = p; _ki = i; _imax = imaxval;
	}

	float	kP() const				{ return _kp; }
	float	kI() const 				{ return _ki; }
	int16_t	imax() const			{ return _imax; }

	void	kP(const float v)		{ _kp=v; }
	void	kI(const float v)		{ _ki=v; }
	void	imax(const int16_t v)	{ _imax=abs(v); }
	float	get_integrator() const	{ return _integrator; }
	void	set_integrator(float i)	{ _integrator = i; }


private:
	float			_kp;
	float			_ki;
	int16_t			_imax;

	float				_integrator;		///< integrator value
};

#endif
